package org.processmining.blocklayout.algorithms;

import java.awt.Font;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jgraph.graph.GraphConstants;
import org.processmining.blocklayout.models.blocks.ElementBlock;
import org.processmining.blocklayout.models.statemachine.Place;
import org.processmining.blocklayout.models.statemachine.StateMachineNet;
import org.processmining.blocklayout.models.statemachine.Transition;
import org.processmining.blocklayout.parameters.LayoutParameters;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;

public class LayoutAlgorithm {

	/**
	 * Lay out the given graph, given the node sizes and additional parameters.
	 * 
	 * @param graph
	 *            The given graph.
	 * @param layout
	 *            The node sizes.
	 * @param layoutParameters
	 *            The additional parameters.
	 */
	public void layout(
			DirectedGraph<? extends DirectedGraphNode, ? extends DirectedGraphEdge<? extends DirectedGraphNode, ? extends DirectedGraphNode>> graph,
			GraphLayoutConnection layout, LayoutParameters layoutParameters) {

		/*
		 * Create a net from the graph.
		 */

		/*
		 * Maps every node in the graph onto a place in the net.
		 */
		Map<DirectedGraphNode, Place> nodes = new HashMap<DirectedGraphNode, Place>();

		/*
		 * Create a state machine net with block structure from the graph.
		 */
		Pair<StateMachineNet, Map<DirectedGraphElement, ElementBlock>> netBlocks = (new StateMachineBuilderAlgorithm()).build(graph, layout, layoutParameters);
		StateMachineNet net = netBlocks.getFirst();
		
		/*
		 * Maps every element in the graph onto its element block. In the end,
		 * this element block will contain the position of this element.
		 */
		Map<DirectedGraphElement, ElementBlock> blocks = netBlocks.getSecond();

//		/*
//		 * Copy all nodes from the graph as places in the net.
//		 */
//		for (DirectedGraphNode node : graph.getNodes()) {
//			/*
//			 * Create a new place for the node.
//			 */
//			Place place = new Place();
//			place.setName(node.getLabel());
//			if (place.getName() == null) {
//				place.setName(node.toString());
//			}
//			/*
//			 * Get the size for the node.
//			 */
//			Dimension dimension = layout.getSize(node);
//			/*
//			 * Create a node block for the place, and set the correct size.
//			 */
//			double height = layoutParameters.getDefaultHeight();
//			double width = layoutParameters.getDefaultWidth();
//			if (dimension != null) {
//				height = dimension.getHeight();
//				width = dimension.getWidth();
//			}
//			if (!layoutParameters.isLeftRight()) {
//				double h = height;
//				height = width;
//				width = h;
//			}
//			NodeBlock block = new NodeBlock(place.getName(), height, width, layoutParameters);
//			place.setBlock(block);
//			/*
//			 * Register the place and the block.
//			 */
//			nodes.put(node, place);
//			blocks.put(node, block);
//		}
//		/*
//		 * Copy all edges from the graph as transitions in the net.
//		 */
//		for (DirectedGraphEdge<? extends DirectedGraphNode, ? extends DirectedGraphNode> edge : graph.getEdges()) {
//			/*
//			 * Create a new transition.
//			 */
//			Transition transition = new Transition();
//			transition.setName(edge.getSource().getLabel() + edge.getTarget().getLabel());
//			/*
//			 * Create a new edge block for this transition.
//			 */
//			EdgeBlock block = new EdgeBlock(transition.getName(), layoutParameters);
//			transition.setBlock(block);
//			/*
//			 * Register the transition and the block.
//			 */
//			blocks.put(edge, block);
//
//			/*
//			 * Add the proper arcs to the net.
//			 */
//			net.addArc(nodes.get(edge.getSource()), transition);
//			net.addArc(transition, nodes.get(edge.getTarget()));
//		}

		/*
		 * Reduce the net, while building up the block structure.
		 */
		layoutParameters.getReductionAlgorithm().reduce(net, layoutParameters);

		/*
		 * Now set the positions of all element blocks in the block structure.
		 */
		double positionY = layoutParameters.getStartY();
		for (Place place : net.getPlaceList()) {
			place.getBlock().setPosition(layoutParameters.getStartX(), positionY, true, layoutParameters);
			positionY += place.getBlock().getTopHeight(layoutParameters)
					+ place.getBlock().getBottomHeight(layoutParameters) + layoutParameters.getVerticalSpace();
		}
		for (Transition transition : net.getTransitionList()) {
			transition.getBlock().setPosition(layoutParameters.getStartX(), positionY, true, layoutParameters);
			positionY += transition.getBlock().getTopHeight(layoutParameters)
					+ transition.getBlock().getBottomHeight(layoutParameters) + layoutParameters.getVerticalSpace();
		}

		/*
		 * Copy the positions from the blocks to the nodes and edges.
		 */
		for (DirectedGraphNode node : graph.getNodes()) {
			/*
			 * Get the position.
			 */
			ElementBlock block = blocks.get(node);
			Point2D position = new Point2D.Double(block.getPositionX(), block.getPositionY());
			if (!layoutParameters.isLeftRight()) {
				position = new Point2D.Double(position.getY(), position.getX());
			}
			System.out.println("[LayoutAlgorithm] Node " + node.getLabel() + " @ " + position);
			/*
			 * Set the position.
			 */
			layout.setPosition(node, position);
		}
		for (DirectedGraphEdge<? extends DirectedGraphNode, ? extends DirectedGraphNode> edge : graph.getEdges()) {

			/*
			 * Get the center position of the source node.
			 */
			ElementBlock sourceBlock = blocks.get(edge.getSource());
			double sourceWidth = sourceBlock.getWidth(layoutParameters);
			Point2D sourcePosition = new Point2D.Double(sourceBlock.getPositionX() + 0.5 * sourceWidth,
					sourceBlock.getPositionY() + sourceBlock.getTopHeight(layoutParameters));
			if (!layoutParameters.isLeftRight()) {
				sourcePosition = new Point2D.Double(sourcePosition.getY(), sourcePosition.getX());
			}

			/*
			 * Get the center position of the target node.
			 */
			ElementBlock targetBlock = blocks.get(edge.getTarget());
			double targetWidth = targetBlock.getWidth(layoutParameters);
			Point2D targetPosition = new Point2D.Double(targetBlock.getPositionX() + 0.5 * targetWidth,
					targetBlock.getPositionY() + targetBlock.getTopHeight(layoutParameters));
			if (!layoutParameters.isLeftRight()) {
				targetPosition = new Point2D.Double(targetPosition.getY(), targetPosition.getX());
			}

			/*
			 * Get the position of the edge.
			 */
			ElementBlock block = blocks.get(edge);
			List<Point2D> edgePoints = new ArrayList<Point2D>();
//			double sourceDist = 1.0 + Math.sqrt(Math.pow(0.5 * sourceWidth, 2.0)
//					+ Math.pow(
//							0.5 * sourceBlock.getTopHeight(layoutParameters)
//									+ sourceBlock.getBottomHeight(layoutParameters), 2.0));
//			double targetDist = 1.0 + Math.sqrt(Math.pow(0.5 * targetWidth, 2.0)
//					+ Math.pow(
//							0.5 * targetBlock.getTopHeight(layoutParameters)
//									+ targetBlock.getBottomHeight(layoutParameters), 2.0));
			Point2D sourcePosition2 = new Point2D.Double(sourceBlock.getWidth(layoutParameters),
					sourceBlock.getTopHeight(layoutParameters) + sourceBlock.getBottomHeight(layoutParameters));
			Point2D targetPosition2 = new Point2D.Double(targetBlock.getWidth(layoutParameters),
					targetBlock.getTopHeight(layoutParameters) + targetBlock.getBottomHeight(layoutParameters));
			if (!layoutParameters.isLeftRight()) {
				sourcePosition2 = new Point2D.Double(sourcePosition2.getY(), sourcePosition2.getX());
				targetPosition2 = new Point2D.Double(targetPosition2.getY(), targetPosition2.getX());
			}
			if (block.getEdgeBlock1() == null || block.getEdgeBlock2() == null) {
				Point2D edgePosition = new Point2D.Double(block.getPositionX(), block.getPositionY());
				if (!layoutParameters.isLeftRight()) {
					edgePosition = new Point2D.Double(edgePosition.getY(), edgePosition.getX());
				}
				System.out.println("[LayoutAlgorithm] Edge " + edge.getSource().getLabel() + ","
						+ edge.getTarget().getLabel() + " @ " + edgePosition);

				/*
				 * Construct the edge points.
				 */
				addEdges(edgePoints, sourcePosition, targetPosition, edgePosition, sourcePosition2, targetPosition2,
						layoutParameters);
			} else {
				Point2D firstPosition = new Point2D.Double(block.getEdgeBlock1().getPositionX(), block.getEdgeBlock1()
						.getPositionY());
				Point2D edgePosition = new Point2D.Double(block.getPositionX(), block.getPositionY());
				Point2D lastPosition = new Point2D.Double(block.getEdgeBlock2().getPositionX(), block.getEdgeBlock2()
						.getPositionY());
				if (!layoutParameters.isLeftRight()) {
					firstPosition = new Point2D.Double(firstPosition.getY(), firstPosition.getX());
					edgePosition = new Point2D.Double(edgePosition.getY(), edgePosition.getX());
					lastPosition = new Point2D.Double(lastPosition.getY(), lastPosition.getX());
				}

				if (Point2D.distance(sourcePosition.getX(), sourcePosition.getY(), firstPosition.getX(),
						firstPosition.getY()) > Point2D.distance(sourcePosition.getX(), sourcePosition.getY(),
						lastPosition.getX(), lastPosition.getY())) {
					lastPosition = new Point2D.Double(block.getEdgeBlock1().getPositionX(), block.getEdgeBlock1()
							.getPositionY());
					firstPosition = new Point2D.Double(block.getEdgeBlock2().getPositionX(), block.getEdgeBlock2()
							.getPositionY());
					if (!layoutParameters.isLeftRight()) {
						firstPosition = new Point2D.Double(firstPosition.getY(), firstPosition.getX());
						lastPosition = new Point2D.Double(lastPosition.getY(), lastPosition.getX());
					}
				}
				System.out.println("[LayoutAlgorithm] Edge " + edge.getSource().getLabel() + ","
						+ edge.getTarget().getLabel() + " @ " + edgePosition);
				Point2D zeroPoint = new Point2D.Double(1.0, 1.0);
				addEdges(edgePoints, sourcePosition, edgePosition, firstPosition, sourcePosition2, zeroPoint,
						layoutParameters);
				addEdges(edgePoints, firstPosition, lastPosition, edgePosition, zeroPoint, zeroPoint, layoutParameters);
				addEdges(edgePoints, edgePosition, targetPosition, lastPosition, zeroPoint, targetPosition2,
						layoutParameters);
			}
			/*
			 * Set the edge points.
			 */
			layout.setEdgePoints(edge, edgePoints);
		}
		for (DirectedGraphNode node : graph.getNodes()) {
			node.getAttributeMap().put(GraphConstants.FONT, new Font("Arial", Font.PLAIN, 30));
		}
	}

	private void addEdges(List<Point2D> edgePoints, Point2D sourcePosition, Point2D targetPosition,
			Point2D edgePosition, Point2D sourcePosition2, Point2D targetPosition2, LayoutParameters layoutParameters) {
		boolean isHorizontal = (sourcePosition.getX() > edgePosition.getX() && edgePosition.getX() > targetPosition.getX())
				|| (sourcePosition.getX() < edgePosition.getX() && edgePosition.getX() < targetPosition.getX());
		if (!layoutParameters.isLeftRight()) {
			isHorizontal = (sourcePosition.getY() > edgePosition.getY() && edgePosition.getY() > targetPosition.getY())
					|| (sourcePosition.getY() < edgePosition.getY() && edgePosition.getY() < targetPosition.getY());
		}
		if (isHorizontal) {
			System.out.println("[LayoutAlgorithm] Hor " + sourcePosition + ", " + edgePosition + ", " + targetPosition);
			Point2D midPoint = getMidPoint(getFracPoint(sourcePosition, edgePosition, layoutParameters.getFraction1()),
					getFracPoint(targetPosition, edgePosition, layoutParameters.getFraction1()));
			Point2D point = getFracPoint(sourcePosition, midPoint, layoutParameters.getFraction2());
			if (doNotOverlap(point, sourcePosition, sourcePosition2, layoutParameters)
					&& doNotOverlap(point, targetPosition, targetPosition2, layoutParameters)) {
				edgePoints.add(point);
			}
			point = getFracPoint(targetPosition, midPoint, layoutParameters.getFraction2());
			if (doNotOverlap(point, sourcePosition, sourcePosition2, layoutParameters)
					&& doNotOverlap(point, targetPosition, targetPosition2, layoutParameters)) {
				edgePoints.add(point);
			}
		} else {
			System.out.println("[LayoutAlgorithm] Ver " + sourcePosition + ", " + edgePosition + ", " + targetPosition);
			double fractionX = layoutParameters.getFraction3();
			double fractionY = layoutParameters.getFraction4();
			if (!layoutParameters.isLeftRight()) {
				fractionX = layoutParameters.getFraction4();
				fractionY = layoutParameters.getFraction3();
			}
			Point2D midPoint = getMidPoint(
					getFracPoint(sourcePosition, edgePosition, fractionX, fractionY),
					getFracPoint(targetPosition, edgePosition, fractionX, fractionY));
			Point2D point = midPoint;
			if (doNotOverlap(point, sourcePosition, sourcePosition2, layoutParameters)
					&& doNotOverlap(point, targetPosition, targetPosition2, layoutParameters)) {
				edgePoints.add(point);
			}
		}
	}

	private boolean doNotOverlap(Point2D point, Point2D midPoint, Point2D dimPoint, LayoutParameters layoutParameters) {
		return (point.getX() + layoutParameters.getEdgePointThreshold() < midPoint.getX() - 0.5 * dimPoint.getX() //
				|| point.getX() - layoutParameters.getEdgePointThreshold() > midPoint.getX() + 0.5 * dimPoint.getX() //
				|| point.getY() + layoutParameters.getEdgePointThreshold() < midPoint.getY() - 0.5 * dimPoint.getY() //
		|| point.getY() - layoutParameters.getEdgePointThreshold() > midPoint.getY() + 0.5 * dimPoint.getY());
	}

//	private void addEdges(List<Point2D> edgePoints, Point2D sourcePosition, Point2D targetPosition,
//			Point2D edgePosition, double sourceDist, double targetDist, LayoutParameters layoutParameters) {
//		if ((sourcePosition.getX() > edgePosition.getX() && edgePosition.getX() > targetPosition.getX())
//				|| (sourcePosition.getX() < edgePosition.getX() && edgePosition.getX() < targetPosition.getX())) {
//			Point2D midPoint = getMidPoint(getFracPoint(sourcePosition, edgePosition, layoutParameters.getFraction1()),
//					getFracPoint(targetPosition, edgePosition, layoutParameters.getFraction1()));
//			Point2D point = getFracPoint(sourcePosition, midPoint, layoutParameters.getFraction2());
//			if (distance(sourcePosition, point) > sourceDist && distance(targetPosition, point) > targetDist) {
//				edgePoints.add(point);
//			}
//			point = getFracPoint(targetPosition, midPoint, layoutParameters.getFraction2());
//			if (distance(sourcePosition, point) > sourceDist && distance(targetPosition, point) > targetDist) {
//				edgePoints.add(point);
//			}
//		} else {
//			Point2D midPoint = getMidPoint(
//					getFracPoint(sourcePosition, edgePosition, layoutParameters.getFraction3(),
//							layoutParameters.getFraction4()),
//					getFracPoint(targetPosition, edgePosition, layoutParameters.getFraction3(),
//							layoutParameters.getFraction4()));
//			Point2D point = midPoint;
//			if (distance(sourcePosition, point) > sourceDist && distance(targetPosition, point) > targetDist) {
//				edgePoints.add(point);
//			}
//		}
//	}

//	private double distance(Point2D pointA, Point2D pointB) {
//		return Math.sqrt(Math.pow(pointA.getX() - pointB.getX(), 2.0) + Math.pow(pointA.getY() - pointB.getY(), 2.0));
//	}

	private Point2D getFracPoint(Point2D pointA, Point2D pointB, double frac) {
		return new Point2D.Double(frac * pointA.getX() + (1.0 - frac) * pointB.getX(), frac * pointA.getY()
				+ (1.0 - frac) * pointB.getY());
	}

	private Point2D getFracPoint(Point2D pointA, Point2D pointB, double fracX, double fracY) {
		return new Point2D.Double(fracX * pointA.getX() + (1.0 - fracX) * pointB.getX(), fracY * pointA.getY()
				+ (1.0 - fracY) * pointB.getY());
	}

	private Point2D getMidPoint(Point2D pointA, Point2D pointB) {
		return getFracPoint(pointA, pointB, 0.5);
	}
}
