package org.processmining.blocklayout.algorithms;

import java.util.HashSet;
import java.util.Set;

import org.processmining.blocklayout.models.reductions.PlaceAbstractionReduction;
import org.processmining.blocklayout.models.reductions.Reduction;
import org.processmining.blocklayout.models.statemachine.Place;
import org.processmining.blocklayout.models.statemachine.StateMachineNet;
import org.processmining.blocklayout.models.statemachine.Transition;
import org.processmining.blocklayout.parameters.LayoutParameters;

public class PlaceAbstractionFinderAlgorithm implements FinderAlgorithm {

	/**
	 * Return the set of suitable reductions for the given net.
	 * 
	 * @param net
	 *            The given net.
	 * @return The set of suitable reductions for the given net.
	 */
	public Set<Reduction> find(StateMachineNet net, LayoutParameters layoutParameters) {
		Set<Reduction> reductions = new HashSet<Reduction>();
		int bestCosts = -1;

		/*
		 * Find an optimal place-transition pair.
		 */
		for (Transition transition : net.getTransitionList()) {
			Place place = net.getPrePlace(transition);
			/*
			 * Place should have a non-empty preset and only the transition in
			 * its postset.
			 * 
			 * Note that transitions in the preset of this place only have this
			 * place as post-place. As a result, these transition cannot have
			 * the post-place of the transition as their post-place.
			 */
			if (net.getPostset(place).size() == 1 && place != net.getPostPlace(transition)) {
				/*
				 * Create the reduction.
				 */
				Reduction reduction = new PlaceAbstractionReduction(place, transition);
				/*
				 * Get the score.
				 */
				int costs = reduction.costs(net, layoutParameters);
				if (costs == 0) {
					/*
					 * Ideal score, remove non-ideal reductions, and stop
					 * looking.
					 */
					reductions.clear();
					reductions.add(reduction);
					return reductions;
				} else if (costs < bestCosts || bestCosts == -1) {
					/*
					 * Found reduction with non-ideal but better score. Remove
					 * lower-scoring reductions, but keep looking.
					 */
					bestCosts = costs;
					reductions.clear();
					reductions.add(reduction);
				}
				/*
				 * Else this reduction has non-best score. Discard.
				 */
			}
		}
		return reductions;
	}
}
