package org.processmining.blocklayout.algorithms;

import java.util.Set;

import org.processmining.blocklayout.models.reductions.Reduction;
import org.processmining.blocklayout.models.statemachine.StateMachineNet;
import org.processmining.blocklayout.parameters.LayoutParameters;

public class ReductionAlgorithm {

	/**
	 * Reduce the given net as far as possible.
	 * 
	 * @param net
	 *            The given net.
	 * @param layoutParameters
	 *            Parameters including, among others, the find algorithm and the
	 *            horizontal space between element blocks.
	 */
	public void reduce(StateMachineNet net, LayoutParameters layoutParameters) {
		while (reduceOnce(net, layoutParameters))
			;
	}

	/**
	 * Try to apply a single reduction on the net.
	 * 
	 * @param net
	 *            The given net.
	 * @param layoutParameters
	 *            Parameters including, among others, the find algorithm and the
	 *            horizontal space between element blocks.
	 * @return Whether a reduction was applied.
	 */
	public boolean reduceOnce(StateMachineNet net, LayoutParameters layoutParameters) {
		int bestCosts = -1;
		Reduction bestReduction = null;
		Reduction reduction = null;
		for (FinderAlgorithm finderAlgorithm : layoutParameters.getFinderAlgorithms()) {
			/*
			 * Find suitable reductions.
			 */
			Set<Reduction> reductions = finderAlgorithm.find(net, layoutParameters);
			if (!reductions.isEmpty()) {
				/*
				 * Suitable reductions found. Take first and score it.
				 */
				reduction = reductions.iterator().next();
				int costs = reduction.costs(net, layoutParameters);
				if (costs == 0) {
					/*
					 * Ideal score. Look no further. Apply it.
					 */
					reduction.apply(net, layoutParameters);
					return true;
				} else if (costs < bestCosts || bestCosts == -1) {
					/*
					 * Non-ideal but better score. New best reduction found.
					 */
					bestReduction = reduction;
					bestCosts = costs;
				}
			}
		}
		/*
		 * No ideal reduction found, try best reduction found.
		 */
		if (bestReduction != null && bestCosts <= layoutParameters.getCostsThreshold()) {
			/*
			 * Best reduction found. Apply it.
			 */
			layoutParameters.getNonZeroCostNets().add(net.createPetriNet(layoutParameters));
			bestReduction.apply(net, layoutParameters);
			return true;
		}
		/*
		 * No reductions found.
		 */
		return false;
	}
}
