package org.processmining.blocklayout.algorithms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import nl.tue.astar.util.ilp.LPMatrix;
import nl.tue.astar.util.ilp.LPMatrixException;

import org.processmining.blocklayout.models.reductions.Reduction;
import org.processmining.blocklayout.models.reductions.TransitionReduction;
import org.processmining.blocklayout.models.statemachine.Place;
import org.processmining.blocklayout.models.statemachine.StateMachineNet;
import org.processmining.blocklayout.models.statemachine.Transition;
import org.processmining.blocklayout.parameters.LayoutParameters;

public class TransitionFinderAlgorithm implements FinderAlgorithm {

	private boolean isReversed;

	public TransitionFinderAlgorithm(boolean isReversed) {
		this.isReversed = isReversed;
	}

	/**
	 * Return the set of suitable reductions for the given net.
	 * 
	 * @param net
	 *            The given net.
	 * @return The set of suitable reductions for the given net.
	 */
	public Set<Reduction> find(StateMachineNet net, LayoutParameters layoutParameters) {
		Set<Reduction> reductions = new HashSet<Reduction>();
		int bestCosts = -1;

		/*
		 * Find a linearly dependent transition.
		 */
		if (net.getTransitions().size() > 1) {
			for (Transition transition : net.getTransitionList()) {
				/*
				 * To be linearly dependent, the pre-place should have
				 * additional post-transitions and the post-place should have
				 * additional pre-transitions.
				 */
				if (net.getPreset(net.getPostPlace(transition)).size() > 1
						&& net.getPostset(net.getPrePlace(transition)).size() > 1) {
					try {
						/*
						 * Create the LP to solve, and try to solve.
						 */
						double[] nativeSolution = solveMatrix(net, transition);
						if (nativeSolution != null) {
							/*
							 * Solution found.
							 */
							List<Transition> transitions = new ArrayList<Transition>();
							int index = 0;
							for (Transition t : net.getTransitionList()) {
								if (nativeSolution[index] > 0.0) {
									transitions.add(t);
								}
								index++;
							}
							/*
							 * Create the reduction.
							 */
							Reduction reduction = new TransitionReduction(transition, transitions, isReversed);
							/*
							 * Get the score of the reduction.
							 */
							int costs = reduction.costs(net, layoutParameters);
							if (costs == 0) {
								/*
								 * Ideal score. Look no further. Remove
								 * non-ideal reductions found so far.
								 */
								reductions.clear();
								reductions.add(reduction);
								return reductions;
							} else if (costs < bestCosts || bestCosts == -1) {
								/*
								 * Non-ideal but better score. Remove reductions
								 * found so far.
								 */
								bestCosts = costs;
								reductions.clear();
								reductions.add(reduction);
							}
							/*
							 * Else non-best score. Discard.
							 */
						}
					} catch (LPMatrixException e) {
					}
				}
			}
		}
		return reductions;
	}

	private double[] solveMatrix(StateMachineNet net, Transition solveTransition) throws LPMatrixException {
//		List<Place> sourcePlaces = net.getSourcePlaces();
//		List<Place> sinkPlaces = net.getSinkPlaces();
		Map<Object, Integer> indexOf = new HashMap<Object, Integer>();
		int index = 0;
		for (Place place : net.getPlaceList()) {
			indexOf.put(place, index);
			index++;
		}
		index = 0;
		for (Transition transition : net.getTransitionList()) {
			indexOf.put(transition, index);
			index++;
		}
		int indexOfSolveTransition = indexOf.get(solveTransition);

		/*
		 * Create incidence matrix
		 */
		int nofRows = net.getPlaces().size();
		int nofCols = net.getTransitions().size(); // + sourcePlaces.size() * sinkPlaces.size();
		int[][] incidenceMatrix = new int[nofRows][nofCols];

		for (int row = 0; row < nofRows; row++) {
			for (int col = 0; col < nofCols; col++) {
				incidenceMatrix[row][col] = 0;
			}
		}

		for (Transition transition : net.getTransitionList()) {
			incidenceMatrix[indexOf.get(net.getPrePlace(transition))][indexOf.get(transition)]--;
			incidenceMatrix[indexOf.get(net.getPostPlace(transition))][indexOf.get(transition)]++;
		}
		int col = net.getTransitions().size();
//		for (Place sourcePlace : sourcePlaces) {
//			for (Place sinkPlace : sinkPlaces) {
//				incidenceMatrix[indexOf.get(sinkPlace)][col]--;
//				incidenceMatrix[indexOf.get(sourcePlace)][col]++;
//				col++;
//			}
//		}

		/*
		 * Create matrix to solve
		 */
		LPMatrix<?> matrix = new LPMatrix.SPARSE.LPSOLVE(nofRows + nofCols, nofCols);

		for (col = 0; col < nofCols; col++) {
			matrix.setObjective(col, 1);
			matrix.adjustMat(col, col, 1.0);
			matrix.setConstrType(col, (col == indexOfSolveTransition ? LPMatrix.EQ : LPMatrix.GE));
			matrix.setRh(col, 0.0);
		}

		for (int row = 0; row < nofRows; row++) {
			for (col = 0; col < nofCols; col++) {
				if (incidenceMatrix[row][col] != 0) {
					matrix.adjustMat(row + nofCols, col, (isReversed && (col == indexOfSolveTransition) ? -1 : 1)
							* incidenceMatrix[row][col]);
				}
			}
			matrix.setConstrType(row + nofCols, LPMatrix.EQ);
			matrix.setRh(row + nofCols, (isReversed ? -1.0 : 1.0) * incidenceMatrix[row][indexOfSolveTransition]);
		}
		matrix.setMinim();

		/*
		 * Solve matrix
		 */
		double[] solution = new double[nofCols];
		if (matrix.solve(solution) == LPMatrix.INFEASIBLE) {
			return null;
		}
		return solution;
	}
}
