package org.processmining.blocklayout.models.blocks;

import java.util.ArrayList;
import java.util.List;

import org.processmining.blocklayout.parameters.LayoutParameters;

public class HorizontalBlock extends AbstractBlock {

	/**
	 * The contained blocks. Always an odd number.
	 */
	private List<Block> blocks;

	private String name;

	/**
	 * Creates a horizontal block from the three given blocks.
	 * 
	 * @param sourceBlock
	 *            The source block. Has connections to the transit block.
	 * @param transitBlock
	 *            The transit block. Has connections from the source block and
	 *            to the target block.
	 * @param targetBlock
	 *            The target block. Has connections from the transit block.
	 */
	public HorizontalBlock(Block sourceBlock, Block transitBlock, Block targetBlock) {
		blocks = new ArrayList<Block>();
		name = null;
		add(sourceBlock);
		add(transitBlock);
		add(targetBlock);
	}

	private void add(Block blockToAdd) {
		if (blockToAdd instanceof HorizontalBlock) {
			HorizontalBlock horizontalBlock = (HorizontalBlock) blockToAdd;
			for (Block block : horizontalBlock.blocks) {
				blocks.add(block);
				if (name == null || name.isEmpty() || (!block.getName().isEmpty() && name.compareTo(block.getName()) > 0)) {
					name = block.getName();
				}
			}
		} else {
			blocks.add(blockToAdd);
			if (name == null || name.isEmpty() || (!blockToAdd.getName().isEmpty() && name.compareTo(blockToAdd.getName()) > 0)) {
				name = blockToAdd.getName();
			}
		}
	}

	/**
	 * Returns the top height of the block.
	 * 
	 * @param layoutParameters
	 *            Parameters containing, among others, the vertical space
	 *            between two blocks.
	 */
	public double getTopHeight(LayoutParameters layoutParameters) {
		/*
		 * The top height of the containing block equals the maximal top height
		 * of any of the contained blocks.
		 */
		double topHeight = 0.0;
		for (Block block : blocks) {
			topHeight = Math.max(topHeight, block.getTopHeight(layoutParameters));
		}
		return topHeight;
	}

	/**
	 * Returns the bottom height of the block.
	 * 
	 * @param layoutParameters
	 *            Parameters containing, among others, the vertical space
	 *            between two blocks.
	 */
	public double getBottomHeight(LayoutParameters layoutParameters) {
		/*
		 * The bottom height of the containing block equals the maximal bottom
		 * height of any of the contained blocks.
		 */
		double bottomHeight = 0.0;
		for (Block block : blocks) {
			bottomHeight = Math.max(bottomHeight, block.getBottomHeight(layoutParameters));
		}
		return bottomHeight;
	}

	/**
	 * Returns the width of the block.
	 * 
	 * @param layoutParameters
	 *            Parameters containing, among others, the horizontal space
	 *            between two blocks.
	 */
	public double getWidth(LayoutParameters layoutParameters) {
		/*
		 * The width of the containing block equals the sum of the widths of the
		 * contained blocks together with the required horizontal spaces between
		 * these blocks.
		 */
		double width = -layoutParameters.getHorizontalSpace();
		for (Block block : blocks) {
			width += block.getWidth(layoutParameters) + layoutParameters.getHorizontalSpace();
		}
		return width;
	}

	/**
	 * Sets the positions of the top-left corners of all contained element
	 * blocks.
	 * 
	 * @param positionX
	 *            The X coordinate of the top-left corner.
	 * @param positionY
	 *            The Y coordinate of the top-left corner.
	 * @param isLeftToRight
	 *            Whether the contained blocks should be laid out from left to
	 *            right (or from right to left) in a horizontal block.
	 * @param layoutParameters
	 *            Parameters containing, among others, the vertical and
	 *            horizontal spaces between element blocks.
	 */
	public void setPosition(double positionX, double positionY, boolean isLeftToRight, LayoutParameters layoutParameters) {
		if (isLeftToRight) {
			for (Block block : blocks) {
				block.setPosition(positionX,
						positionY + getTopHeight(layoutParameters) - block.getTopHeight(layoutParameters),
						isLeftToRight, layoutParameters);
				positionX += block.getWidth(layoutParameters) + layoutParameters.getHorizontalSpace();
			}
		} else {
			List<Block> reverseBlocks = new ArrayList<Block>();
			for (Block block : blocks) {
				reverseBlocks.add(0, block);
			}
			for (Block block : reverseBlocks) {
				block.setPosition(positionX,
						positionY + getTopHeight(layoutParameters) - block.getTopHeight(layoutParameters),
						isLeftToRight, layoutParameters);
				positionX += block.getWidth(layoutParameters) + layoutParameters.getHorizontalSpace();
			}
		}
	}

	public int getNofLayers() {
		int nofLayers = 0;
		for (Block block : blocks) {
			nofLayers += block.getNofLayers();
		}
		return nofLayers;
	}

	public String getName() {
		return name;
	}
}
