package org.processmining.blocklayout.models.blocks;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.processmining.blocklayout.parameters.LayoutParameters;

public class VerticalBlock extends AbstractBlock {

	/**
	 * The contained blocks. Always an odd number.
	 */
	private List<Block> blocks;
	private List<Block> reverseBlocks;

	private String name;

	private static Comparator<Block> blockComparator = new Comparator<Block>() {

		public int compare(Block b1, Block b2) {
			if (b1.getName().isEmpty()) {
				return b2.getName().length();
			} else if (b2.getName().isEmpty()) {
				return -b1.getName().length();
			}
			return b1.getName().compareTo(b2.getName());
		}

	};

	/**
	 * Creates a vertical block from the two given blocks.
	 * 
	 * @param topBlock
	 *            The top block.
	 * @param bottomBlock
	 *            The bottom block.
	 * @param bottomReverse
	 *            Whether the direction of the bottom block is the reverse of
	 *            the direction of the top block.
	 */
	public VerticalBlock(Block topBlock, Block bottomBlock, boolean bottomReverse) {
		blocks = new ArrayList<Block>();
		reverseBlocks = new ArrayList<Block>();
		name = null;
		add(topBlock, false);
		add(bottomBlock, bottomReverse);
	}

	private void add(Block blockToAdd, boolean isReverse) {
		if (blockToAdd instanceof VerticalBlock) {
			VerticalBlock verticalBlock = (VerticalBlock) blockToAdd;
			for (Block block : isReverse ? verticalBlock.reverseBlocks : verticalBlock.blocks) {
				blocks.add(block);
				if (name == null || name.isEmpty() || (!block.getName().isEmpty() && name.compareTo(block.getName()) > 0)) {
					name = block.getName();
				}
			}
			for (Block block : isReverse ? verticalBlock.blocks : verticalBlock.reverseBlocks) {
				reverseBlocks.add(block);
				if (name == null || name.isEmpty() || (!block.getName().isEmpty() && name.compareTo(block.getName()) > 0)) {
					name = block.getName();
				}
			}
		} else {
			if (isReverse) {
				reverseBlocks.add(blockToAdd);
			} else {
				blocks.add(blockToAdd);
			}
			if (name == null || name.isEmpty() || (!blockToAdd.getName().isEmpty() && name.compareTo(blockToAdd.getName()) > 0)) {
				name = blockToAdd.getName();
			}
		}

		Collections.sort(blocks, blockComparator);
//		System.out.println("[VerticalBlock] ---");
		for (Block block : blocks) {
//			System.out.println("[VerticalBlock] " + block.getName());
		}
		Collections.sort(reverseBlocks, blockComparator);
//		System.out.println("[VerticalBlock] ---");
		for (Block block : reverseBlocks) {
//			System.out.println("[VerticalBlock] " + block.getName());
		}
//		System.out.println("[VerticalBlock] ---");
	}

	/**
	 * Returns the top height of the block.
	 * 
	 * @param layoutParameters
	 *            Parameters containing, among others, the vertical space
	 *            between element blocks.
	 */
	public double getTopHeight(LayoutParameters layoutParameters) {
		double topHeight = -layoutParameters.getVerticalSpace();
		for (Block block : blocks) {
			topHeight += layoutParameters.getVerticalSpace() + block.getTopHeight(layoutParameters)
					+ block.getBottomHeight(layoutParameters);
		}
		topHeight *= 0.5;
		return topHeight;
	}

	/**
	 * Returns the bottom height of the block.
	 * 
	 * @param layoutParameters
	 *            Parameters containing, among others, the vertical space
	 *            between element blocks.
	 */
	public double getBottomHeight(LayoutParameters layoutParameters) {
		double bottomHeight = getTopHeight(layoutParameters);
		for (Block block : reverseBlocks) {
			bottomHeight += layoutParameters.getVerticalSpace() + block.getTopHeight(layoutParameters)
					+ block.getBottomHeight(layoutParameters);
		}
		return bottomHeight;
	}

	/**
	 * Returns the width of the block.
	 * 
	 * @param layoutParameters
	 *            Parameters containing, among others, the horizontal space
	 *            between element blocks.
	 */
	public double getWidth(LayoutParameters layoutParameters) {
		double width = 0.0;
		for (Block block : blocks) {
			width = Math.max(width, block.getWidth(layoutParameters));
		}
		for (Block block : reverseBlocks) {
			width = Math.max(width, block.getWidth(layoutParameters));
		}
		return width;
	}

	/**
	 * Sets the positions of the top-left corners of all contained element
	 * blocks.
	 * 
	 * @param positionX
	 *            The X coordinate of the top-left corner.
	 * @param positionY
	 *            The Y coordinate of the top-left corner.
	 * @param isLeftToRight
	 *            Whether the contained block should be laid out from left to
	 *            right (or from right to left) in a horizontal block.
	 * @param layoutParameters
	 *            Parameters containing, among others, the vertical and
	 *            horizontal spaces between element blocks.
	 */
	public void setPosition(double positionX, double positionY, boolean isLeftToRight, LayoutParameters layoutParameters) {
		/*
		 * Top block is topmost block.
		 */
		double horizontalSpace = layoutParameters.getHorizontalSpace();
		int nofLayers = getNofLayers();
		for (Block block : blocks) {
			int nofBlockLayers = (nofLayers == 1 ? 1 : block.getNofLayers());
			if (nofBlockLayers > 1 || nofLayers == 1) {
				layoutParameters.setHorizontalSpace(horizontalSpace
						+ ((getWidth(layoutParameters) - block.getWidth(layoutParameters)) / (nofBlockLayers - 1)));
				block.setPosition(positionX, positionY, isLeftToRight, layoutParameters);
				layoutParameters.setHorizontalSpace(horizontalSpace);
			} else {
				EdgeBlock edgeBlock1 = new EdgeBlock("", layoutParameters);
				EdgeBlock edgeBlock2 = new EdgeBlock("", layoutParameters);
				HorizontalBlock horizontalBlock = new HorizontalBlock(edgeBlock1, block, edgeBlock2);
				nofBlockLayers = horizontalBlock.getNofLayers();
				layoutParameters.setHorizontalSpace(horizontalSpace
						+ ((getWidth(layoutParameters) - horizontalBlock.getWidth(layoutParameters)) / (nofBlockLayers - 1)));
				horizontalBlock.setPosition(positionX, positionY, isLeftToRight, layoutParameters);
				layoutParameters.setHorizontalSpace(horizontalSpace);
				((ElementBlock) block).setEdgeBlocks(edgeBlock1, edgeBlock2);
//				block.setPosition(positionX + 0.5 * (getWidth(layoutParameters) - block.getWidth(layoutParameters)),
//						positionY, isLeftToRight, layoutParameters);
			}
			positionY += block.getTopHeight(layoutParameters) + block.getBottomHeight(layoutParameters)
					+ layoutParameters.getVerticalSpace();
		}
		for (Block block : reverseBlocks) {
			int nofBlockLayers = (nofLayers == 1 ? 1 : block.getNofLayers());
			if (nofBlockLayers > 1 || nofLayers == 1) {
				layoutParameters.setHorizontalSpace(horizontalSpace
						+ ((getWidth(layoutParameters) - block.getWidth(layoutParameters)) / (nofBlockLayers - 1)));
				block.setPosition(positionX, positionY, !isLeftToRight, layoutParameters);
				layoutParameters.setHorizontalSpace(horizontalSpace);
			} else {
				EdgeBlock edgeBlock1 = new EdgeBlock("", layoutParameters);
				EdgeBlock edgeBlock2 = new EdgeBlock("", layoutParameters);
				HorizontalBlock horizontalBlock = new HorizontalBlock(edgeBlock1, block, edgeBlock2);
				nofBlockLayers = horizontalBlock.getNofLayers();
				layoutParameters.setHorizontalSpace(horizontalSpace
						+ ((getWidth(layoutParameters) - horizontalBlock.getWidth(layoutParameters)) / (nofBlockLayers - 1)));
				horizontalBlock.setPosition(positionX, positionY, isLeftToRight, layoutParameters);
				layoutParameters.setHorizontalSpace(horizontalSpace);
				((ElementBlock) block).setEdgeBlocks(edgeBlock1, edgeBlock2);
//				block.setPosition(positionX + 0.5 * (getWidth(layoutParameters) - block.getWidth(layoutParameters)),
//						positionY, !isLeftToRight, layoutParameters);
			}
			positionY += block.getTopHeight(layoutParameters) + block.getBottomHeight(layoutParameters)
					+ layoutParameters.getVerticalSpace();
		}
	}

	public int getNofLayers() {
		int nofLayers = 0;
		for (Block block : blocks) {
			nofLayers = Math.max(nofLayers, block.getNofLayers());
		}
		for (Block block : reverseBlocks) {
			nofLayers = Math.max(nofLayers, block.getNofLayers());
		}
		return nofLayers;
	}

	public String getName() {
		return name;
	}
}
