package org.processmining.blocklayout.models.reductions;

import org.processmining.blocklayout.models.blocks.Block;
import org.processmining.blocklayout.models.blocks.HorizontalBlock;
import org.processmining.blocklayout.models.statemachine.Place;
import org.processmining.blocklayout.models.statemachine.StateMachineNet;
import org.processmining.blocklayout.models.statemachine.Transition;
import org.processmining.blocklayout.parameters.LayoutParameters;

public class TransitionAbstractionReduction implements Reduction {

	/**
	 * The place involved in this reduction.
	 */
	private Place place;

	/**
	 * The transition involve din this reduction.
	 */
	private Transition transition;

	private int costs;

	/**
	 * Creates the reduction for the given place and the given transition.
	 * 
	 * @param place
	 *            The given place.
	 * @param transition
	 *            The given transition.
	 */
	public TransitionAbstractionReduction(Place place, Transition transition) {
		this.place = place;
		this.transition = transition;
		this.costs = -1;
	}

	/**
	 * Returns the costs of applying this reduction on the given net.
	 * 
	 * @param net
	 *            The given net
	 * @return The costs of applying this reduction on the given net.
	 */
	public int costs(StateMachineNet net, LayoutParameters layoutParameters) {
		if (costs == -1) {
			if (net.getPreset(place).isEmpty()) {
				costs = Integer.MAX_VALUE;
			} else if (net.getPreset(place).size() == 1) {
				costs = 0;
			} else {
				costs = layoutParameters.getTransitionAbstractionReductionCosts(net, place, transition);
			}
			System.out.println("[TransitionAbstractionReduction] Costs " + costs + " ("
					+ place + ") + [" + transition + "]");
		}
		return costs;
	}

	/**
	 * Applies this reduction to the given net.
	 * 
	 * @param net
	 *            The given net.
	 * @param layoutParameters
	 *            Parameters including, among others, the horizontal space
	 *            between element blocks.
	 */
	public void apply(StateMachineNet net, LayoutParameters layoutParameters) {
		/*
		 * We merge the given place, the given transition, and one of the
		 * transitions from the preset of the given place into a single
		 * transition.
		 */
		Transition bestTransition = net.getPreset(place).iterator().next();
		/*
		 * Select the transition with maximal width, as this minimizes the
		 * chance that any unselected transition ends either above or below the
		 * given place and given transition in the end.
		 */
		for (Transition t : net.getPreset(place)) {
			if (t.getBlock().getWidth(layoutParameters) > bestTransition.getBlock().getWidth(layoutParameters)) {
				bestTransition = t;
			}
		}
		System.out.println("[TransitionAbstractionReduction] Assign " + costs(net, layoutParameters) + " ["
				+ bestTransition + "] <- (" + place + ") + [" + transition + "]");
		Block block = new HorizontalBlock(bestTransition.getBlock(), place.getBlock(), transition.getBlock());
		bestTransition.setBlock(block);
		/*
		 * Remove the given place and given transition from the net.
		 */
		net.applyAbstraction(place, transition);
	}

}
