package org.processmining.blocklayout.models.reductions;

import java.util.ArrayList;
import java.util.List;

import org.processmining.blocklayout.models.blocks.Block;
import org.processmining.blocklayout.models.blocks.VerticalBlock;
import org.processmining.blocklayout.models.statemachine.Place;
import org.processmining.blocklayout.models.statemachine.StateMachineNet;
import org.processmining.blocklayout.models.statemachine.Transition;
import org.processmining.blocklayout.parameters.LayoutParameters;

public class TransitionReduction implements Reduction {

	/**
	 * The linearly dependent transition.
	 */
	private Transition transition;

	/**
	 * The transitions the transition is linearly dependent on.
	 */
	private List<Transition> transitions;

	/**
	 * Whether or not to reverse the arcs for the linearly dependent transition.
	 * If reversed, we're looking for a loop back.
	 */
	private boolean isReverse;

	private int costs = -1;

	/**
	 * Create the transition reduction for the given linearly dependent
	 * transition, the set of transitions the transition is linearly dependent
	 * on, and whether the arcs to th elinearly dependent transition were
	 * reversed.
	 * 
	 * @param transition
	 *            The linearly dependent transition.
	 * @param transitions
	 *            The transitions the transition is linearly dependent on.
	 * @param isReverse
	 *            Whether or not to reverse the arcs for the linearly dependent
	 *            transition.
	 */
	public TransitionReduction(Transition transition, List<Transition> transitions, boolean isReverse) {
		this.transition = transition;
		this.transitions = new ArrayList<Transition>(transitions);
		this.isReverse = isReverse;
		this.costs = -1;
	}

	/**
	 * Returns the costs of applying this reduction on the given net.
	 * 
	 * @param net
	 *            The given net.
	 * @return The costs of applying the reduction on the given net.
	 */
	public int costs(StateMachineNet net, LayoutParameters layoutParameters) {
		if (costs == -1) {
			if (transitions.size() == 1) {
				costs = 0;
			} else {
				costs = layoutParameters.getTransitionReductionCosts(net, transition, transitions);
			}
			System.out.println("[TransitionReduction] Costs " + costs + " [" + transition + "]");
		}
		return costs;
	}

	/**
	 * Applies this reduction to the given net.
	 * 
	 * @param net
	 *            The given net.
	 * @param layoutParameters
	 *            Parameters including, among others, the horizontal space
	 *            between element blocks.
	 */
	public void apply(StateMachineNet net, LayoutParameters layoutParameters) {
		if (transitions.isEmpty()) {
			/*
			 * Transition linearly dependent on empty set, must be self-loop
			 * transition.
			 */
			Place place = net.getPrePlace(transition);
			System.out.println("[TransitionReduction] Assign false " + costs(net, layoutParameters) + " [" + transition
					+ "] -> (" + place + ")");
			Block block = new VerticalBlock(place.getBlock(), transition.getBlock(), true);
			place.setBlock(block);
		} else {
			/*
			 * Find the best transition from the set of transitions. For the
			 * time being, we will take a transition with maximal width as the
			 * best.
			 */
			Transition bestTransition = transitions.iterator().next();
			for (Transition t : transitions) {
				if (t.getBlock().getWidth(layoutParameters) > bestTransition.getBlock().getWidth(layoutParameters)) {
					bestTransition = t;
				}
			}
			System.out.println("[TransitionReduction] Assign " + isReverse + " " + costs(net, layoutParameters) + " ["
					+ transition + "] -> [" + bestTransition + "]");
			Block block = new VerticalBlock(bestTransition.getBlock(), transition.getBlock(), isReverse);
			bestTransition.setBlock(block);
		}
		/*
		 * Remove the given transition from the net.
		 */
		net.applyTransition(transition);
	}

}
