package org.processmining.blocklayout.models.statemachine;

import java.awt.Dimension;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.blocklayout.parameters.LayoutParameters;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;

public class StateMachineNet {

	/**
	 * Presets and postsets of places.
	 */
	private Map<Place, List<Transition>> preTransitions;
	private Map<Place, List<Transition>> postTransitions;

	/**
	 * Pre-place and post-place of transitions.
	 */
	private Map<Transition, Place> prePlace;
	private Map<Transition, Place> postPlace;

	/**
	 * Ordered lists of transitions and places.
	 */
	private List<Transition> transitionList;
	private List<Place> placeList;

	/**
	 * Unordered sets of transitions and places.
	 */
	private Set<Transition> transitions;
	private Set<Place> places;

	/**
	 * Creates an empty net.
	 */
	public StateMachineNet() {
		preTransitions = new HashMap<Place, List<Transition>>();
		postTransitions = new HashMap<Place, List<Transition>>();
		prePlace = new HashMap<Transition, Place>();
		postPlace = new HashMap<Transition, Place>();
		transitionList = new ArrayList<Transition>();
		placeList = new ArrayList<Place>();
		transitions = new HashSet<Transition>();
		places = new HashSet<Place>();
	}

	public Petrinet createPetriNet(LayoutParameters layoutParameters) {
		Petrinet pn = new PetrinetImpl("State Machine net with " + transitions.size() + " transitions");

		Map<Place, org.processmining.models.graphbased.directed.petrinet.elements.Place> placeMap = new HashMap<Place, org.processmining.models.graphbased.directed.petrinet.elements.Place>();
		for (Place place : placeList) {
			org.processmining.models.graphbased.directed.petrinet.elements.Place pnPlace = pn.addPlace(place.getName());
			placeMap.put(place, pnPlace);
			int height = Math.max(2, (int) (place.getBlock().getTopHeight(layoutParameters)
					+ place.getBlock().getBottomHeight(layoutParameters)));
			int width = Math.max(2, (int) place.getBlock().getWidth(layoutParameters));
			pnPlace.getAttributeMap().put(AttributeMap.SIZE, new Dimension(width, height));
		}
		for (Transition transition : transitionList) {
			org.processmining.models.graphbased.directed.petrinet.elements.Transition pnTransition = pn
					.addTransition(transition.getName());
			pn.addArc(placeMap.get(prePlace.get(transition)), pnTransition);
			pn.addArc(pnTransition, placeMap.get(postPlace.get(transition)));
			int height = Math.max(2, (int) (transition.getBlock().getTopHeight(layoutParameters)
					+ transition.getBlock().getBottomHeight(layoutParameters)));
			int width = Math.max(2, (int) transition.getBlock().getWidth(layoutParameters));
			pnTransition.getAttributeMap().put(AttributeMap.SIZE, new Dimension(width, height));
		}
		return pn;
	}

	/**
	 * Adds an arc from the place to the transition to the net. Also adds the
	 * place and/or transition if needed.
	 * 
	 * @param place
	 *            The source place of the arc.
	 * @param transition
	 *            The target transition of the arc.
	 */
	public void addArc(Place place, Transition transition) {
		/*
		 * Add the place to the net, if needed.
		 */
		addPlace(place);
		/*
		 * Add the transition to the net, if needed.
		 */
		addTransition(transition);
		/*
		 * Set the place as the pre-place of the transition.
		 */
		prePlace.put(transition, place);
		/*
		 * Add the transition as a post-transition of the place.
		 */
		postTransitions.get(place).add(transition);
	}

	/**
	 * Adds an arc from the transition to the place to the net. Also adds the
	 * place and/or transition if needed.
	 * 
	 * @param transition
	 *            The source transition of the arc.
	 * @param place
	 *            The target place of the arc.
	 */
	public void addArc(Transition transition, Place place) {
		/*
		 * Add the place to the net, if needed.
		 */
		addPlace(place);
		/*
		 * Add the transition to the net, if needed.
		 */
		addTransition(transition);
		/*
		 * Set the place as the post-place of the transition.
		 */
		postPlace.put(transition, place);
		/*
		 * Add the transition as a pre-transition of the place.
		 */
		preTransitions.get(place).add(transition);
	}

	/**
	 * Adds a place to the net, if needed.
	 * 
	 * @param place
	 *            The place to add.
	 */
	public void addPlace(Place place) {
		if (places.add(place)) {
			/*
			 * Is needed. Add place.
			 */
			placeList.add(place);
			/*
			 * Initialize preset and postset.
			 */
			postTransitions.put(place, new ArrayList<Transition>());
			preTransitions.put(place, new ArrayList<Transition>());
		}
	}

	/**
	 * Adds a transition to the net, if needed.
	 * 
	 * @param transition
	 *            The transition to add.
	 */
	public void addTransition(Transition transition) {
		if (transitions.add(transition)) {
			/*
			 * Is needed. Add transition.
			 */
			transitionList.add(transition);
			/*
			 * No need to initialize pre-place and post-place. Will be set
			 * later.
			 */
		}
	}

	/*
	 * Getters.
	 */

	public List<Transition> getPreset(Place place) {
		return preTransitions.get(place);
	}

	public List<Transition> getPostset(Place place) {
		return postTransitions.get(place);
	}

	public Place getPrePlace(Transition transition) {
		return prePlace.get(transition);
	}

	public Place getPostPlace(Transition transition) {
		return postPlace.get(transition);
	}

	/**
	 * Applies the abstraction rule on the given place and transition in the
	 * net.
	 * 
	 * @param place
	 *            The given place.
	 * @param transition
	 *            The given transition.
	 */
	public void applyAbstraction(Place place, Transition transition) {
		/*
		 * Adds the new arcs from any transition in the preset of the preset of
		 * the given place to the only place in the postset of the given
		 * transition.
		 * 
		 * As a result, the given place will not be a post-place for these
		 * transitions any more.
		 */
		for (Transition t : preTransitions.get(place)) {
			addArc(t, postPlace.get(transition));
		}
		/*
		 * Remove the transition from the preset of its post-place.
		 */
		preTransitions.get(postPlace.get(transition)).remove(transition);
		/*
		 * The given place and transition are now detached. As a result, we can
		 * now remove them.
		 */
		preTransitions.keySet().remove(place);
		postTransitions.keySet().remove(place);
		prePlace.keySet().remove(transition);
		postPlace.keySet().remove(transition);
		places.remove(place);
		placeList.remove(place);
		transitions.remove(transition);
		transitionList.remove(transition);
	}

	/**
	 * Applies the transition rule on the given transition.
	 * 
	 * @param transition
	 *            The given transition.
	 */
	public void applyTransition(Transition transition) {
		/*
		 * Remove the transition from the preset of the post-place and postset
		 * of the pre-place.
		 */
		postTransitions.get(prePlace.get(transition)).remove(transition);
		preTransitions.get(postPlace.get(transition)).remove(transition);
		/*
		 * The transition is now detached. As a result, we can remove it.
		 */
		prePlace.keySet().remove(transition);
		postPlace.keySet().remove(transition);
		transitions.remove(transition);
		transitionList.remove(transition);
	}

	public void applyPlace(Place place, Place proxy) {
		for (Transition transition : preTransitions.get(place)) {
			postPlace.put(transition, proxy);
		}
		for (Transition transition : postTransitions.get(place)) {
			prePlace.put(transition, proxy);
		}
		preTransitions.keySet().remove(place);
		postTransitions.keySet().remove(place);
		places.remove(place);
		placeList.remove(place);
	}

	public void applyRigid(Transition splitTransition, Transition transition, Transition joinTransition) {
		// Detach the transition.
		postTransitions.get(prePlace.get(transition)).remove(transition);
		preTransitions.get(postPlace.get(transition)).remove(transition);
		// Connect the transition to split and join places. 
		Place splitPlace = getPrePlace(splitTransition);
		Place joinPlace = getPostPlace(joinTransition);
		prePlace.put(transition, splitPlace);
		postPlace.put(transition, joinPlace);
		postTransitions.get(splitPlace).add(transition);
		preTransitions.get(joinPlace).add(transition);
		// This removes the rigid. 
	}
	
	/*
	 * More getters.
	 */

	public Set<Transition> getTransitions() {
		return transitions;
	}

	public List<Transition> getTransitionList() {
		return transitionList;
	}

	public Set<Place> getPlaces() {
		return places;
	}

	public List<Place> getPlaceList() {
		return placeList;
	}

	/**
	 * Returns an ordered list containing all source places.
	 * 
	 * @return The ordered list containing all source places.
	 */
	public List<Place> getSourcePlaces() {
		List<Place> sourcePlaces = new ArrayList<Place>(places);
		sourcePlaces.removeAll(postPlace.values());
		return sourcePlaces;
	}

	/**
	 * Returns an ordered list containing all sink places.
	 * 
	 * @return The ordered list containing all sink places.
	 */
	public List<Place> getSinkPlaces() {
		List<Place> sinkPlaces = new ArrayList<Place>(places);
		sinkPlaces.removeAll(prePlace.values());
		return sinkPlaces;

	}
}
