package org.processmining.blocklayout.parameters;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.blocklayout.algorithms.AlternativeTransitionFinderAlgorithm;
import org.processmining.blocklayout.algorithms.FinderAlgorithm;
import org.processmining.blocklayout.algorithms.PlaceAbstractionFinderAlgorithm;
import org.processmining.blocklayout.algorithms.PlaceFinderAlgorithm;
import org.processmining.blocklayout.algorithms.ReductionAlgorithm;
import org.processmining.blocklayout.algorithms.ReverseTransitionFinderAlgorithm;
import org.processmining.blocklayout.algorithms.RigidFinderAlgorithm;
import org.processmining.blocklayout.algorithms.SelfLoopTransitionFinderAlgorithm;
import org.processmining.blocklayout.algorithms.TransitionAbstractionFinderAlgorithm;
import org.processmining.blocklayout.algorithms.TransitionFinderAlgorithm;
import org.processmining.blocklayout.models.statemachine.Place;
import org.processmining.blocklayout.models.statemachine.StateMachineNet;
import org.processmining.blocklayout.models.statemachine.Transition;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;

public class LayoutParameters {

	private double horizontalSpace;
	private double verticalSpace;
	private double defaultHeight;
	private double defaultWidth;
	private double startX;
	private double startY;
	private double fraction1;
	private double fraction2;
	private double fraction3;
	private double fraction4;
	private boolean leftRight;
	private double edgePointThreshold;
	private ReductionAlgorithm reductionAlgorithm;
	private List<FinderAlgorithm> finderAlgorithms;
	private List<Petrinet> nonZeroCostNets;
	private int costsThreshold;

	public LayoutParameters() {
		finderAlgorithms = new ArrayList<FinderAlgorithm>();
		nonZeroCostNets = new ArrayList<Petrinet>();
		
		reductionAlgorithm = new ReductionAlgorithm();

		// Quick-but-incomplete transition rules
		finderAlgorithms.add(new AlternativeTransitionFinderAlgorithm());
		finderAlgorithms.add(new SelfLoopTransitionFinderAlgorithm());
		finderAlgorithms.add(new ReverseTransitionFinderAlgorithm());
		// Abstraction rules
		finderAlgorithms.add(new TransitionAbstractionFinderAlgorithm());
		finderAlgorithms.add(new PlaceAbstractionFinderAlgorithm());
		// Slow-but-complete transition rules
		finderAlgorithms.add(new TransitionFinderAlgorithm(false));
		finderAlgorithms.add(new TransitionFinderAlgorithm(true));
		// Place rule
		finderAlgorithms.add(new PlaceFinderAlgorithm(false));
		finderAlgorithms.add(new PlaceFinderAlgorithm(true));
		// Rigid rule
		finderAlgorithms.add(new RigidFinderAlgorithm());
		
		setHorizontalSpace(10.0);
		setVerticalSpace(20.0);
		setDefaultWidth(50.0);
		setDefaultHeight(40.0);
		setStartX(10.0);
		setStartY(10.0);
		setFraction1(0.1);
		setFraction2(0.25);
		setFraction3(0.0);
		setFraction4(1.1);
		setLeftRight(true);
		setEdgePointThreshold(5.0);
		setCostsThreshold(Integer.MAX_VALUE);
	}

	public double getHorizontalSpace() {
		return horizontalSpace;
	}

	public void setHorizontalSpace(double horizontalSpace) {
		this.horizontalSpace = horizontalSpace;
	}

	public double getVerticalSpace() {
		return verticalSpace;
	}

	public void setVerticalSpace(double verticalSpace) {
		this.verticalSpace = verticalSpace;
	}

	public double getDefaultHeight() {
		return defaultHeight;
	}

	public void setDefaultHeight(double defaultHeight) {
		this.defaultHeight = defaultHeight;
	}

	public double getDefaultWidth() {
		return defaultWidth;
	}

	public void setDefaultWidth(double defaultWidth) {
		this.defaultWidth = defaultWidth;
	}

	public ReductionAlgorithm getReductionAlgorithm() {
		return reductionAlgorithm;
	}

	public void setReductionAlgorithm(ReductionAlgorithm reductionAlgorithm) {
		this.reductionAlgorithm = reductionAlgorithm;
	}

	public List<FinderAlgorithm> getFinderAlgorithms() {
		return finderAlgorithms;
	}
	
	public List<Petrinet> getNonZeroCostNets() {
		return nonZeroCostNets;
	}

	public double getFraction1() {
		return fraction1;
	}

	public void setFraction1(double fraction1) {
		this.fraction1 = fraction1;
	}

	public double getFraction2() {
		return fraction2;
	}

	public void setFraction2(double fraction2) {
		this.fraction2 = fraction2;
	}

	public double getFraction3() {
		return fraction3;
	}

	public void setFraction3(double fraction3) {
		this.fraction3 = fraction3;
	}

	public double getFraction4() {
		return fraction4;
	}

	public void setFraction4(double fraction4) {
		this.fraction4 = fraction4;
	}

	public double getStartX() {
		return startX;
	}

	public void setStartX(double startX) {
		this.startX = startX;
	}

	public double getStartY() {
		return startY;
	}

	public void setStartY(double startY) {
		this.startY = startY;
	}

	public boolean isLeftRight() {
		return leftRight;
	}

	public void setLeftRight(boolean leftRight) {
		this.leftRight = leftRight;
	}

	public double getEdgePointThreshold() {
		return edgePointThreshold;
	}

	public void setEdgePointThreshold(double edgePointThreshold) {
		this.edgePointThreshold = edgePointThreshold;
	}

	/*
	 * By overriding the following default cost functions, the actual layout can be influenced.
	 * Note that these methods are not called if the costs are either zero or infinite. 
	 */
	
	/*
	 * Returns the costs when the provided place and transition get reduced by
	 * abstraction into the output place of the provided transition.
	 */
	public int getPlaceAbstractionReductionCosts(StateMachineNet net, Place place, Transition transition) {
		return 2 * (net.getPreset(net.getPostPlace(transition)).size() - 1);
	}

	/*
	 * Returns the costs when the provided place and transition get reduced by
	 * abstraction into one of the input transitions of the provided place.
	 */
	public int getTransitionAbstractionReductionCosts(StateMachineNet net, Place place, Transition transition) {
		return 4 * (net.getPreset(place).size() - 1);
	}

	/*
	 * Returns the costs when the provided place and proxy place get merged.
	 */
	public int getPlaceReductionCosts(StateMachineNet net, Place place, Place proxy) {
		Set<Place> preset1 = new HashSet<Place>();
		Set<Place> preset2 = new HashSet<Place>();
		Set<Place> postset1 = new HashSet<Place>();
		Set<Place> postset2 = new HashSet<Place>();

		for (Transition transition : net.getPreset(place)) {
			preset1.add(net.getPrePlace(transition));
		}
		for (Transition transition : net.getPreset(proxy)) {
			preset2.add(net.getPrePlace(transition));
		}
		for (Transition transition : net.getPostset(place)) {
			postset1.add(net.getPostPlace(transition));
		}
		for (Transition transition : net.getPostset(proxy)) {
			postset2.add(net.getPostPlace(transition));
		}

		Set<Place> preUnion = new HashSet<Place>(preset1);
		preUnion.addAll(preset2);
		Set<Place> preDiff = new HashSet<Place>(preset1);
		preDiff.retainAll(preset2);
		Set<Place> postUnion = new HashSet<Place>(postset1);
		postUnion.addAll(postset2);
		Set<Place> postDiff = new HashSet<Place>(postset1);
		postDiff.retainAll(postset2);
		return (preUnion.size() - preDiff.size()) + (postUnion.size() - postDiff.size());
	}

	/*
	 * Returns the costs when the provided transition gets reduction using the
	 * transition rule where the provided collection of transitions are the
	 * transitions the provided transition is linearly dependent on.
	 */
	public int getTransitionReductionCosts(StateMachineNet net, Transition transition, List<Transition> transitions) {
		if (transitions.isEmpty()) {
			return 0;
		}
		return transitions.size() - 1;
	}

	public int getCostsThreshold() {
		return costsThreshold;
	}

	public void setCostsThreshold(int costsThreshold) {
		this.costsThreshold = costsThreshold;
	}
}
